"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const tag_expressions_1 = __importDefault(require("@cucumber/tag-expressions"));
const uuid_1 = require("uuid");
const assertions_1 = require("./assertions");
const data_table_1 = __importDefault(require("./data_table"));
const registry_1 = require("./registry");
const ast_helpers_1 = require("./ast-helpers");
const constants_1 = require("./constants");
const INTERNAL_PROPERTY_NAME = "_cypress-cucumber-preprocessor-do-not-use";
/**
 * From messages.TestStepFinished.TestStepResult.Status.
 */
const Status = {
    Unknown: "UNKNOWN",
    Passed: "PASSED",
    Skipped: "SKIPPED",
    Pending: "PENDING",
    Undefined: "UNDEFINED",
    Ambiguous: "AMBIGUOUS",
    Failed: "FAILED",
};
function storeInternalProperties(properties) {
    cy.wrap(properties, { log: false }).as(INTERNAL_PROPERTY_NAME);
}
function retrieveInternalProperties() {
    return cy.get("@" + INTERNAL_PROPERTY_NAME, {
        log: false,
    });
}
function createFeature(context, feature) {
    describe(feature.name || "<unamed feature>", () => {
        if (feature.children) {
            for (const child of feature.children) {
                if (child.scenario) {
                    createScenario(context, child.scenario);
                }
                else if (child.rule) {
                    createRule(context, child.rule);
                }
            }
        }
    });
}
function createRule(context, rule) {
    describe(rule.name || "<unamed rule>", () => {
        if (rule.children) {
            for (const child of rule.children) {
                if (child.scenario) {
                    createScenario(context, child.scenario);
                }
            }
        }
    });
}
function createWeakCache(mapper) {
    return {
        cache: new WeakMap(),
        get(key) {
            const cacheHit = this.cache.get(key);
            if (cacheHit) {
                return cacheHit;
            }
            const value = mapper(key);
            this.cache.set(key, value);
            return value;
        },
    };
}
const gherkinDocumentsAstIdMaps = createWeakCache((key) => {
    const astIdMap = new Map();
    for (const node of (0, ast_helpers_1.traverseGherkinDocument)(key)) {
        if ("id" in node && node.id) {
            astIdMap.set(node.id, node);
        }
    }
    return astIdMap;
});
function createScenario(context, scenario) {
    if (scenario.examples) {
        const exampleIds = scenario.examples
            .map((examples) => {
            return (0, assertions_1.assertAndReturn)(examples.tableBody, "Expected to find a table body").map((row) => (0, assertions_1.assertAndReturn)(row.id, "Expected table row to have an id"));
        })
            .reduce((acum, el) => acum.concat(el), []);
        for (let i = 0; i < exampleIds.length; i++) {
            const exampleId = exampleIds[i];
            const pickle = (0, assertions_1.assertAndReturn)(context.pickles.find((pickle) => pickle.astNodeIds && pickle.astNodeIds.includes(exampleId)), `Expected to find a pickle associated with id = ${exampleId}`);
            const baseName = (0, assertions_1.assertAndReturn)(pickle.name, "Expected pickle to have a name");
            const exampleName = `${baseName} (example #${i + 1})`;
            createPickle(context, Object.assign(Object.assign({}, scenario), { name: exampleName }), pickle);
        }
    }
    else {
        const scenarioId = (0, assertions_1.assertAndReturn)(scenario.id, "Expected scenario to have an id");
        const pickle = (0, assertions_1.assertAndReturn)(context.pickles.find((pickle) => pickle.astNodeIds && pickle.astNodeIds.includes(scenarioId)), `Expected to find a pickle associated with id = ${scenarioId}`);
        createPickle(context, scenario, pickle);
    }
}
function createPickle(context, scenario, pickle) {
    var _a;
    const { registry, gherkinDocument, pickles, testFilter, messages } = context;
    const testCaseId = (0, uuid_1.v4)();
    const definitionIds = [];
    if (pickle.steps) {
        for (const step of pickle.steps) {
            const id = (0, uuid_1.v4)();
            definitionIds.push(id);
            messages.stack.push({
                stepDefinition: {
                    id,
                    pattern: {
                        source: "a step",
                        type: "CUCUMBER_EXPRESSION",
                    },
                    sourceReference: {
                        uri: "not available",
                        location: { line: 0 },
                    },
                },
            });
        }
    }
    const testCase = {
        testCase: {
            id: testCaseId,
            pickleId: pickle.id,
            testSteps: (_a = pickle.steps) === null || _a === void 0 ? void 0 : _a.map((step, i) => {
                return {
                    id: step.id,
                    pickleStepId: step.id,
                    stepDefinitionIds: [definitionIds[i]],
                };
            }),
        },
    };
    messages.stack.push(testCase);
    const astIdMap = gherkinDocumentsAstIdMaps.get(gherkinDocument);
    const tags = pickle.tags ? pickle.tags.map(ast_helpers_1.mapTagName) : [];
    if (!testFilter.evaluate(tags)) {
        return;
    }
    let attempt = 0;
    it(scenario.name || "<unamed scenario>", { env: { tags } }, function () {
        var _a, _b, _c, _d, _e;
        (0, registry_1.assignRegistry)(registry);
        const testCaseStartedId = (0, uuid_1.v4)();
        const testCaseStarted = {
            testCaseStarted: {
                id: testCaseStartedId,
                testCaseId,
                attempt: attempt++,
            },
        };
        messages.stack.push(testCaseStarted);
        window.testState = {
            gherkinDocument,
            pickles,
            pickle,
        };
        if (!pickle.steps) {
            storeInternalProperties({
                pickle,
                testCaseStartedId,
                remainingSteps: [],
            });
        }
        if (pickle.steps) {
            const remainingSteps = [...pickle.steps];
            storeInternalProperties({
                pickle,
                testCaseStartedId,
                remainingSteps,
            });
            for (const pickleStep of pickle.steps) {
                const text = (0, assertions_1.assertAndReturn)(pickleStep.text, "Expected pickle step to have a text");
                const scenarioStep = (0, assertions_1.assertAndReturn)(astIdMap.get((0, assertions_1.assertAndReturn)((_a = pickleStep.astNodeIds) === null || _a === void 0 ? void 0 : _a[0], "Expected to find at least one astNodeId")), `Expected to find scenario step associated with id = ${(_b = pickleStep.astNodeIds) === null || _b === void 0 ? void 0 : _b[0]}`);
                cy.wrap(null, { log: false }).then(() => {
                    Cypress.log({
                        name: "step",
                        displayName: (0, assertions_1.assertAndReturn)("keyword" in scenarioStep && scenarioStep.keyword, "Expected to find a keyword in the scenario step"),
                        message: text, // This property was wrongfully typed in 3.x.y.
                    });
                });
                const argument = ((_c = pickleStep.argument) === null || _c === void 0 ? void 0 : _c.dataTable)
                    ? new data_table_1.default(pickleStep.argument.dataTable)
                    : ((_e = (_d = pickleStep.argument) === null || _d === void 0 ? void 0 : _d.docString) === null || _e === void 0 ? void 0 : _e.content)
                        ? pickleStep.argument.docString.content
                        : undefined;
                cy.then(() => {
                    const testStepStarted = {
                        testStepStarted: {
                            testStepId: pickleStep.id,
                            testCaseStartedId,
                        },
                    };
                    messages.stack.push(testStepStarted);
                    if (messages.enabled) {
                        cy.task(constants_1.TASK_TEST_STEP_STARTED, pickleStep.id, { log: false });
                    }
                })
                    .then(() => registry.runStepDefininition(this, text, argument))
                    .then((result) => {
                    if (result === "pending") {
                        const testStepFinished = {
                            testStepFinished: {
                                testStepId: pickleStep.id,
                                testCaseStartedId,
                                testStepResult: {
                                    status: Status.Pending,
                                },
                            },
                        };
                        messages.stack.push(testStepFinished);
                        remainingSteps.shift();
                        for (const skippedStep of remainingSteps) {
                            const skippedTestStepStarted = {
                                testStepStarted: {
                                    testStepId: skippedStep.id,
                                    testCaseStartedId,
                                },
                            };
                            messages.stack.push(skippedTestStepStarted);
                            const skippedTestStepFinished = {
                                testStepFinished: {
                                    testStepId: skippedStep.id,
                                    testCaseStartedId,
                                    testStepResult: {
                                        status: Status.Skipped,
                                    },
                                },
                            };
                            messages.stack.push(skippedTestStepFinished);
                        }
                        for (let i = 0, count = remainingSteps.length; i < count; i++) {
                            remainingSteps.pop();
                        }
                        this.skip();
                    }
                    else {
                        const testStepFinished = {
                            testStepFinished: {
                                testStepId: pickleStep.id,
                                testCaseStartedId,
                                testStepResult: {
                                    status: Status.Passed,
                                },
                            },
                        };
                        messages.stack.push(testStepFinished);
                        remainingSteps.shift();
                    }
                });
            }
        }
    });
}
function collectTagNamesFromGherkinDocument(gherkinDocument) {
    const tagNames = [];
    for (const node of (0, ast_helpers_1.traverseGherkinDocument)(gherkinDocument)) {
        if ("tags" in node && node.tags) {
            for (const tag of node.tags) {
                tagNames.push((0, ast_helpers_1.mapTagName)(tag));
            }
        }
    }
    return tagNames;
}
function createTests(registry, source, gherkinDocument, pickles, messagesEnabled) {
    const noopNode = { evaluate: () => true };
    const environmentTags = Cypress.env("TAGS");
    const messages = [];
    messages.push({
        source: { data: source, uri: gherkinDocument.uri },
    });
    messages.push({
        gherkinDocument: Object.assign({}, gherkinDocument),
    });
    for (const pickle of pickles) {
        messages.push({
            pickle,
        });
    }
    const testFilter = collectTagNamesFromGherkinDocument(gherkinDocument).includes("@focus")
        ? (0, tag_expressions_1.default)("@focus")
        : environmentTags
            ? (0, tag_expressions_1.default)(environmentTags)
            : noopNode;
    if (gherkinDocument.feature) {
        createFeature({
            registry,
            gherkinDocument,
            pickles,
            testFilter,
            messages: {
                enabled: messagesEnabled,
                stack: messages,
            },
        }, gherkinDocument.feature);
    }
    afterEach(function () {
        (0, registry_1.freeRegistry)();
        retrieveInternalProperties().then((properties) => {
            var _a, _b, _c, _d;
            const { testCaseStartedId, remainingSteps } = properties;
            if (remainingSteps.length > 0) {
                const error = (0, assertions_1.assertAndReturn)((_b = (_a = this.currentTest) === null || _a === void 0 ? void 0 : _a.err) === null || _b === void 0 ? void 0 : _b.message, "Expected to find an error message");
                const failedStep = (0, assertions_1.assertAndReturn)(remainingSteps.shift(), "Expected there to be a remaining step");
                const failedTestStepFinished = error.includes("Step implementation missing")
                    ? {
                        testStepFinished: {
                            testStepId: failedStep.id,
                            testCaseStartedId,
                            testStepResult: {
                                status: Status.Undefined,
                            },
                        },
                    }
                    : {
                        testStepFinished: {
                            testStepId: failedStep.id,
                            testCaseStartedId,
                            testStepResult: {
                                status: Status.Failed,
                                message: (_d = (_c = this.currentTest) === null || _c === void 0 ? void 0 : _c.err) === null || _d === void 0 ? void 0 : _d.message,
                            },
                        },
                    };
                messages.push(failedTestStepFinished);
                for (const skippedStep of remainingSteps) {
                    const skippedTestStepStarted = {
                        testStepStarted: {
                            testStepId: skippedStep.id,
                            testCaseStartedId,
                        },
                    };
                    messages.push(skippedTestStepStarted);
                    const skippedTestStepFinished = {
                        testStepFinished: {
                            testStepId: skippedStep.id,
                            testCaseStartedId,
                            testStepResult: {
                                status: Status.Skipped,
                            },
                        },
                    };
                    messages.push(skippedTestStepFinished);
                }
            }
            const testCaseFinished = {
                testCaseFinished: {
                    testCaseStartedId,
                },
            };
            messages.push(testCaseFinished);
        });
    });
    after(() => {
        if (messagesEnabled) {
            cy.task(constants_1.TASK_APPEND_MESSAGES, messages, { log: false });
        }
    });
}
exports.default = createTests;
