"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getRegistry = exports.freeRegistry = exports.assignRegistry = exports.withRegistry = exports.Registry = void 0;
const cucumber_expressions_1 = require("@cucumber/cucumber-expressions");
const tag_expressions_1 = __importDefault(require("@cucumber/tag-expressions"));
const assertions_1 = require("./assertions");
const type_guards_1 = require("./type-guards");
const noopNode = { evaluate: () => true };
function retrieveCurrentTagsOrWarn() {
    const isFeatureSpec = Cypress.spec.name.endsWith(".feature");
    const envTags = Cypress.env("tags");
    const tags = isFeatureSpec
        ? envTags
        : envTags !== null && envTags !== void 0 ? envTags : (cy.log("⚠️ cypress-cucumber-preprocessor: Using Cucumber hooks with non-feature files is supported and you can tag them yourself using the `tags` environment variable. Use an explicit, empty array to remove this warning."),
            []);
    if (!Array.isArray(tags)) {
        throw new Error("cypress-cucumber-preprocessor: Expected an array of tags. This might mean that you have erroneously overriden the environment variable yourself.");
    }
    else if (!tags.every(type_guards_1.isString)) {
        throw new Error("cypress-cucumber-preprocessor: Expected an array of strings. This might mean that you have erroneously overriden the environment variable yourself.");
    }
    else {
        return tags;
    }
}
class Registry {
    constructor() {
        this.preliminaryStepDefinitions = [];
        this.defineStep = this.defineStep.bind(this);
        this.runStepDefininition = this.runStepDefininition.bind(this);
        this.defineParameterType = this.defineParameterType.bind(this);
        this.defineBefore = this.defineBefore.bind(this);
        this.defineAfter = this.defineAfter.bind(this);
        this.parameterTypeRegistry = new cucumber_expressions_1.ParameterTypeRegistry();
        this.stepDefinitions = [];
    }
    finalize() {
        for (const { description, implementation } of this
            .preliminaryStepDefinitions) {
            if (typeof description === "string") {
                this.stepDefinitions.push({
                    expression: new cucumber_expressions_1.CucumberExpression(description, this.parameterTypeRegistry),
                    implementation,
                });
            }
            else {
                this.stepDefinitions.push({
                    expression: new cucumber_expressions_1.RegularExpression(description, this.parameterTypeRegistry),
                    implementation,
                });
            }
        }
    }
    defineStep(description, implementation) {
        if (typeof description !== "string" && !(description instanceof RegExp)) {
            throw new Error("Unexpected argument for step definition");
        }
        this.preliminaryStepDefinitions.push({
            description,
            implementation,
        });
    }
    defineParameterType({ name, regexp, transformer, }) {
        this.parameterTypeRegistry.defineParameterType(new cucumber_expressions_1.ParameterType(name, regexp, null, transformer, true, false));
    }
    defineBefore(options, fn) {
        const node = options.tags ? (0, tag_expressions_1.default)(options.tags) : noopNode;
        beforeEach(function () {
            if (node.evaluate(retrieveCurrentTagsOrWarn())) {
                fn.call(this);
            }
        });
    }
    defineAfter(options, fn) {
        const node = options.tags ? (0, tag_expressions_1.default)(options.tags) : noopNode;
        afterEach(function () {
            if (node.evaluate(retrieveCurrentTagsOrWarn())) {
                fn.call(this);
            }
        });
    }
    resolveStepDefintion(text) {
        const matchingStepDefinitions = this.stepDefinitions.filter((stepDefinition) => stepDefinition.expression.match(text));
        if (matchingStepDefinitions.length === 0) {
            throw new Error(`Step implementation missing for: ${text}`);
        }
        else if (matchingStepDefinitions.length > 1) {
            throw new Error(`Multiple matching step definitions for: ${text}\n` +
                matchingStepDefinitions
                    .map((stepDefinition) => {
                    const { expression } = stepDefinition;
                    if (expression instanceof cucumber_expressions_1.RegularExpression) {
                        return ` ${expression.regexp}`;
                    }
                    else if (expression instanceof cucumber_expressions_1.CucumberExpression) {
                        return ` ${expression.source}`;
                    }
                })
                    .join("\n"));
        }
        else {
            return matchingStepDefinitions[0];
        }
    }
    runStepDefininition(world, text, argument) {
        const stepDefinition = this.resolveStepDefintion(text);
        const args = stepDefinition.expression
            .match(text)
            .map((match) => match.getValue(world));
        if (argument) {
            args.push(argument);
        }
        return stepDefinition.implementation.apply(world, args);
    }
}
exports.Registry = Registry;
const globalPropertyName = "__cypress_cucumber_preprocessor_registry_dont_use_this";
function withRegistry(fn) {
    const registry = new Registry();
    assignRegistry(registry);
    fn();
    freeRegistry();
    return registry;
}
exports.withRegistry = withRegistry;
function assignRegistry(registry) {
    globalThis[globalPropertyName] = registry;
}
exports.assignRegistry = assignRegistry;
function freeRegistry() {
    delete globalThis[globalPropertyName];
}
exports.freeRegistry = freeRegistry;
function getRegistry() {
    return (0, assertions_1.assertAndReturn)(globalThis[globalPropertyName], "Expected to find a global registry (this usually means you are trying to define steps or hooks in support/index.js, which is not supported)");
}
exports.getRegistry = getRegistry;
